@@ -1,12 +1,17 @@
 package Gtk2::Ex::HyperTextBuffer;
 
-our $VERSION = '0.16';
+our $VERSION = '0.18';
 
 use strict;
 use Gtk2;
 use Gtk2::Pango; # pango constants
 use File::Spec;  # for locating images
-use Glib::Object::Subclass Gtk2::TextBuffer:: ;
+use Glib::Object::Subclass
+	Gtk2::TextBuffer::,
+	signals => {
+		# new signals
+		edit_mode_changed => { },
+	} ;
 
 =head1 NAME
 
@@ -27,9 +32,7 @@ instead of plain text. It also keeps an undo-/redo-stack.
 
 =cut
 
-# It seems to me that 10 * PANGO_SCALE is the normal font size
 # TODO make tag styles configable
-# FIXME replace "weight" with "scale" ?
 our %TAGS = (
 	bold      => [weight => PANGO_WEIGHT_BOLD],
 	italic    => [style => 'italic'],
@@ -38,11 +41,11 @@ our %TAGS = (
 	verbatim  => [family => 'monospace'], # wrap_mode => 'none'], # Bug withwrap_mode !?
 	Verbatim  => [family => 'monospace', wrap_mode => 'none', indent => 40],
 	link      => [foreground => 'blue', underline => 'single'],
-	head1     => [weight => PANGO_WEIGHT_BOLD, size => 10*1.2**3 * PANGO_SCALE ],
-	head2     => [weight => PANGO_WEIGHT_BOLD, size => 10*1.2**2 * PANGO_SCALE ],
-	head3     => [weight => PANGO_WEIGHT_BOLD, size => 12 * PANGO_SCALE        ],
-	head4     => [weight => PANGO_WEIGHT_BOLD, size => 10 * PANGO_SCALE        ],
-	head5     => [style  => 'italic', weight => PANGO_WEIGHT_BOLD, size => 10 * PANGO_SCALE],
+	head1     => [weight => PANGO_WEIGHT_BOLD, scale => 1.15**4 ],
+	head2     => [weight => PANGO_WEIGHT_BOLD, scale => 1.15**3 ],
+	head3     => [weight => PANGO_WEIGHT_BOLD, scale => 1.15**2 ],
+	head4     => [weight => PANGO_WEIGHT_ULTRABOLD, scale => 1.15 ],
+	head5     => [weight => PANGO_WEIGHT_BOLD, scale => 1.15, style  => 'italic'],
 );
 
 our @_tags = grep {$_ ne 'link'} keys %TAGS;
@@ -71,14 +74,15 @@ Simple constructor.
 sub INIT_INSTANCE {
 	my $self = shift;
 	$self->{edit_mode_tags} = [];
-	$self->{image_marks} = [];
-#	$self->signal_connect(mark_set => sub {
-#			my ($self, $iter, $mark) = @_;
-#			return unless $mark->get_name eq 'insert';
-#			my @names = map {$_->get_property('name')} $self->_get_open_tags($iter);
-#			warn "Tags: @names\n";
-#			#$self->set_edit_mode_from_cursor;
-#		} );
+	$self->{undo} = [];
+	$self->{redo} = [];
+	$self->signal_connect(mark_set => sub {
+			my ($self, $iter, $mark) = @_;
+			$self->set_edit_mode_from_cursor()
+				if $mark->get_name eq 'insert';
+		} );
+	$self->signal_connect_after(insert_text => \&on_insert);
+	$self->signal_connect(delete_range => \&on_delete_range);
 }
 
 =item C<clear()>
@@ -86,7 +90,7 @@ sub INIT_INSTANCE {
 Deletes all buffer contents.
 
 Be aware that this module sometimes stores a reference to the buffer in a
-global variable in order to get copy-pasting with mark up work. This means
+global variable in order to get copy-pasting with markup work. This means
 you could save a lot of memory when you always clear the object when you are
 done with it.
 
@@ -95,7 +99,6 @@ done with it.
 sub clear {
 	my $self = shift;
 	$self->{edit_mode_tags} = [];
-	$self->{image_marks} = [];
 	$self->delete( $self->get_bounds );
 }
 
@@ -116,7 +119,7 @@ the format as used by L<Pod::Simple> but it uses more verbose element names.
 
 For example:
 
-	['Document', {},
+	['Page', {},
 		['head1', {}, 'Test page'],
 		"\n\n",
 		"This is a test page showing some syntax:\n\n",
@@ -127,28 +130,40 @@ For example:
 		], "\n",
 	]
 
+This function creates a new unmodified buffer.
+This action is not un-doable.
+
 =cut
 
 sub set_parse_tree {
 	my ($self, $tree) = @_;
 	#use Data::Dumper; print "HyperTextBuffer got tree: ", Dumper $tree;
 	$self->clear;
+	$self->{_undo_lock} = 1;
 	$self->insert_blocks_at_cursor($tree);
+	$self->{_undo_lock} = 0;
 	$self->set_modified(0);
 }
 
 =item C<insert_blocks(ITER, BLOCK, ...)>
 
 Insert one or more partial parse trees in the buffer at ITER.
+The blocks will also be added to the undo stack.
 
 =cut
 
 sub insert_blocks {
 	my $self = shift;
 	my $iter = shift;
+	my $insert = $self->get_insert;
+	my $mark = $self->create_mark(
+		'insert-blocks-orig-insert',
+		$self->get_iter_at_mark($insert),
+		1 );
 	$self->place_cursor($iter);
-	$self->set_edit_mode_from_cursor;
 	$self->insert_blocks_at_cursor(@_);
+	$self->place_cursor( $self->get_iter_at_mark($mark) );
+	$self->delete_mark($mark);
 }
 
 =item C<insert_blocks_at_cursor(BLOCK, ...)>
@@ -162,14 +177,22 @@ my $_Verbatim = 0;
 sub insert_blocks_at_cursor {
 	# Make sure not to modify content of the reference !
 	# TODO: check default editable / region editable
+	my $self = shift;
+	
+	$self->signal_handlers_block_by_func(\&on_insert);
+	my $begin = $self->get_iter_at_mark( $self->get_insert )->get_offset;
+	$self->_insert_blocks(@_);
+	my $end = $self->get_iter_at_mark( $self->get_insert )->get_offset;
+	$self->_record_step('I', $begin, $end, @_);
+	$self->signal_handlers_unblock_by_func(\&on_insert);
+	return 1;
+}
+
+sub _insert_blocks {
 	my ($self, @blocks) = @_;
-	#use Data::Dumper; warn Dumper \@blocks;
-#	$self->set_edit_mode_from_cursor
-#		unless defined $self->{edit_mode_tags}; # FIXME
-	$self->{edit_mode_tags} ||= [];
+
 	for (@blocks) {
 		unless (ref $_) {
-			$_ =~ s/^(\s*)\*(\s+)/$1\x{2022}$2/mg unless $_Verbatim; # FIXME doesn't belong in this module !
 			if (@{$self->{edit_mode_tags}}) {
 				my $iter = $self->get_iter_at_mark( $self->get_insert );
 				$self->insert_with_tags(
@@ -181,7 +204,7 @@ sub insert_blocks_at_cursor {
 			my ($type, $meta) = @$_[0, 1];
 			my $tag;
 			if ($type eq 'image') {
-				$self->insert_image_from_file($$meta{src});
+				$self->insert_image_from_file(@$meta{qw/src width height/});
 				next;
 			}
 			elsif ($type eq 'link') {
@@ -193,17 +216,17 @@ sub insert_blocks_at_cursor {
 			if ($tag) {
 				$_Verbatim++ if $type eq 'Verbatim'; # FIXME
 				push @{$self->{edit_mode_tags}}, $tag;
-				$self->insert_blocks_at_cursor(@$_[2 .. $#$_]); # recurs
+				$self->_insert_blocks(@$_[2 .. $#$_]); # recurs
 				pop @{$self->{edit_mode_tags}};
 				$_Verbatim-- if $type eq 'Verbatim'; # FIXME
 			}
-			elsif ($type eq 'Document' or $type eq 'Para') {
+			elsif ($type eq 'Page' or $type eq 'Para') {
 				if ($$meta{base}) {
 					$self->{base} = $$meta{base};
 					$self->{base} =~ s#^file://##;
 					#warn "base set to: $self->{base}\n";
 				}
-				$self->insert_blocks_at_cursor(@$_[2 .. $#$_]); # recurs
+				$self->_insert_blocks(@$_[2 .. $#$_]); # recurs
 			}
 			else { die "Unknown tag in parse tree: $type" }
 
@@ -213,36 +236,53 @@ sub insert_blocks_at_cursor {
 		}
 	}
 
-	return 1;
 }
 
-=item C<insert_image_from_file(FILENAME)>
+=item C<insert_image_from_file(FILENAME, WIDTH, HEIGHT, ITER)>
 
-Insert an image in the buffer at the cursor.
+Insert an image in the buffer at the cursor
+or at ITER if ITER is specified.
+
+WIDTH and HEIGHT are optional arguments. If both are set the image
+is resized to these dimensions. If one of both is zero or undefined
+the image is scaled with fixed ratio.
+
+Use C<insert_blocks()> if you want this action to be un-doable. 
 
 =cut
 
 sub insert_image_from_file {
-	my ($self, $src) = @_;
+	my ($self, $src, $w, $h, $iter) = @_;
+	($w, $iter) = (undef, $w) if @_ == 3 and ref $w;
 	my $file = $self->{base} ? File::Spec->rel2abs($src, $self->{base}) : $src;
 	#warn "file: $src, base: $self->{base} => $file\n";
 
 	my $pixbuf = undef;
-	eval { $pixbuf = Gtk2::Gdk::Pixbuf->new_from_file($file) }
-		if (-f $file and -r _) ;
+	if (-f $file and -r _) {
+		unless ($w || $h) {
+			eval { $pixbuf =
+				Gtk2::Gdk::Pixbuf->new_from_file($file) }
+		}
+		else {
+			eval { $pixbuf =
+				Gtk2::Gdk::Pixbuf->new_from_file_at_size(
+					$file, $w || -1, $h || -1) }
+		}
+	}
 	$pixbuf ||= Gtk2::Image->new->render_icon('gtk-missing-image', 'dialog')->copy;
 	die "Could not insert image: $file\n" unless $pixbuf;
 	
 	$pixbuf->{image_src} = $src;
 	$pixbuf->{image_file} = $file;
-	my $iter = $self->get_iter_at_mark( $self->get_insert );
+	$pixbuf->{resized} = ($w || $h) ? [$w, $h] : 0;
+	$iter ||= $self->get_iter_at_mark( $self->get_insert );
 
 	#my $mark = $self->create_mark(undef, $iter, 0);
 	#$mark->set_visible(1);
 	
 	$self->insert_pixbuf($iter, $pixbuf);
 
-	return 1;
+	return $pixbuf;
 }
 
 =item C<get_parse_tree()>
@@ -278,7 +318,6 @@ sub get_parse_tree {
 		
 		EXTRACT_TAGS:
 		my $slice = $previous->get_slice($iter);
-		$slice =~ s/^(\s*)\x{2022}(\s+)/$1\*$2/mg; # FIXME totally wrong level
 		if ($slice =~ /\x{fffc}/) {
 			my $image = $previous->copy;
 			while ($slice =~ s/^(.*?)\x{fffc}//s) { # Embeded image
@@ -286,8 +325,10 @@ sub get_parse_tree {
 				$image->forward_chars(length $1);
 				my $pixbuf = $image->get_pixbuf;
 				if ($pixbuf) {
-					my $src = $pixbuf->{image_src};
-					push @{$stack[-1]}, ['image', {src => $src}];
+					my $meta = { src => $pixbuf->{image_src} };
+					@$meta{qw/width height/} = @{$pixbuf->{resized}}
+						if $pixbuf->{resized};
+					push @{$stack[-1]}, ['image', $meta];
 				}
 				$image->forward_char;
 			}
@@ -332,7 +373,6 @@ sub get_parse_tree {
 			$$next[2] =~ s/^(\n*)//;
 			$lines += length $1;
 			splice @$next, 2, 1 unless length $$next[2];
-			$$root[$_+1] = undef unless @{$$root[$_+1]} > 2;
 		}
 		elsif (defined $$root[$_+1]) {
 			$$root[$_+1] =~ s/^(\n*)//;
@@ -355,10 +395,17 @@ sub get_parse_tree {
 		
 		$$root[$_][1]{empty_lines} = $lines - 1 if $lines > 1;
 	}
-	@$root = grep defined($_), @$root;
+
+	# Filter out empty tags
+	@$root = grep {
+		defined($_) and
+		(ref($_) and ($$_[0] !~ /^(link|image)/)) 
+			? (@$_ > 2)
+			: length($_)
+	} @$root;
 		
-	unshift @$root, 'Document', {};
-	#use Data::Dumper; print STDERR Dumper $root;
+	unshift @$root, 'Page', {};
+	#use Data::Dumper; warn "get_parse_tree: ", Dumper $root;
 	return $root;
 }
 
@@ -447,6 +494,7 @@ sub select_line {
 	$begin = $self->get_iter_at_line( $begin->get_line )
 		unless $begin->starts_line;
 	while (! $iter->ends_line) { $iter->forward_char }
+	return unless $iter->compare($begin) == 1;
 	$self->select_range($begin, $iter);
 	return ($begin, $iter);
 }
@@ -491,17 +539,6 @@ sub select_link {
 	}
 }
 
-=item C<apply_tag_to_selection>
-
-TODO
-
-=cut
-
-sub apply_tag_to_selection {
-	my ($self, $tag) = @_;
-	
-}
-
 =item C<replace_selection(TEXT)>
 
 Replace any selected text with TEXT.
@@ -526,9 +563,10 @@ sub replace_selection {
 	return $start, $end;
 }
 
-=item C<set_edit_mode_tags>
+=item C<set_edit_mode_tags(TAG, ..)>
 
-TODO
+Set the current edit mode to a certain list of tags. This means that all
+text you insert will get these tags applied.
 
 =cut
 
@@ -536,15 +574,31 @@ sub set_edit_mode_tags {
 	my $self = shift;
 	my $table = $self->get_tag_table;
 	$self->{edit_mode_tags} =
-		[ map { ref($_) ? $_ : $table->lookup($_) } @$_ ];
+		[ map { ref($_) ? $_ : $table->lookup($_) } @_ ];
+	$self->signal_emit('edit_mode_changed');
 }
 
-=item C<set_edit_mode_from_cursor>
+=item C<set_edit_mode_from_cursor()>
 
-TODO
+Like C<set_edit_mode_tags()> but copies tags from the cursor position.
+This function is called automaticly after moving the cursor.
 
 =cut
 
+sub set_edit_mode_from_cursor {
+	my $self = shift;
+	my $iter = $self->get_iter_at_mark( $self->get_insert );
+	$self->{edit_mode_tags} = [$self->_get_open_tags($iter)];
+	my $string = "@{$self->{edit_mode_tags}}";
+	my $changed = ($string ne $self->{_edit_mode_string});
+	$self->{_edit_mode_string} = $string;
+	$self->signal_emit('edit_mode_changed') if $changed;
+	#warn join ' ', "Mode:", map(
+	#	{$_->{is_link} ? 'link' : $_->get_property('name')}
+	#	@{$self->{edit_mode_tags}}
+	#), "\n" if $changed;
+}
+
 sub _get_open_tags {
 	# iter->get_tags has "left" gravity
 	# this function has "right" gravity
@@ -554,21 +608,13 @@ sub _get_open_tags {
 		@tags = grep {$_ ne $t} @tags;
 	}
 	push @tags, $iter->get_toggled_tags(0);
+	@tags = grep {$_->get_property('name') !~ /gtkspell/} @tags;
 	return @tags;
 }
 
-sub set_edit_mode_from_cursor {
-	my $self = shift;
-	my $iter = $self->get_iter_at_mark( $self->get_insert );
-	$self->{edit_mode_tags} = [$iter->get_tags];
-	#print join ' ', "Mode:", map(
-	#	{$_->get_property('name')} @{$self->{edit_mode_tags}}
-	#), "\n";
-}
+=item C<get_edit_mode_tags()>
 
-=item C<get_edit_mode_tags>
-
-TODO
+Returns a list of tags that will be applied to newly entered text.
 
 =cut
 
@@ -578,49 +624,95 @@ sub get_edit_mode_tags {
 	@{$_[0]->{edit_mode_tags}};
 }
 
-=item C<toggle_edit_mode_tag>
+=item C<toggle_edit_mode_tag(TAG, BOOL)>
 
-TODO
+Toggles a tag for editing. If BOOL is undefined it just toggles, else it sets the tag to the value of BOOL.
 
 =cut
 
 sub toggle_edit_mode_tag {
-	my ($self, $tag) = @_;
+	my ($self, $tag, $bool) = @_;
 	$tag = $self->get_tag_table->lookup($tag) unless ref $tag;
-	$self->set_edit_mode_from_cursor
-		unless defined $self->{edit_mode_tags};
-	if (grep {$_ eq $tag} @{$self->{edit_mode_tags}}) {
-		@{$self->{edit_mode_tags}} = grep {$_ ne $tag} @{$self->{edit_mode_tags}};
-	}
-	else {
-		push @{$self->{edit_mode_tags}}, $tag;
-	}
+	my @tags = grep {$_ ne $tag} @{$self->{edit_mode_tags}};
+	push @tags, $tag
+		if @tags == @{$self->{edit_mode_tags}} and $bool != 0;
+	@{$self->{edit_mode_tags}} = @tags;
+	$self->signal_emit('edit_mode_changed');
 }
 
-=item C<reset_edit_mode>
+sub on_insert {
+	my ($self, $end, $string) = @_;
+	# length in argument list does not understand utf8
+	my $start = $end->copy;
+	$start->backward_chars(length $string);
+	$self->remove_all_tags($start, $end);
+	$self->apply_tag($_, $start, $end) for @{$self->{edit_mode_tags}};
+	$self->_record_step('I', $start->get_offset, $end->get_offset, $string);
+}
 
-TODO
-
-=cut
-
-sub reset_edit_mode { $_[0]->{edit_mode_tags} = undef }
+sub on_delete_range {
+	my ($self, $begin, $end) = @_;
+	my $tree = $self->get_parse_tree($begin, $end);
+	$self->_record_step('D', $begin->get_offset, $end->get_offset, $tree);
+}
 
 =item C<undo>
 
 TODO
 
-=cut
-
-sub undo {
-}
-
 =item C<redo>
 
 TODO
 
 =cut
 
+my %reverse_step = (
+	I => 'D',
+	D => 'I',
+	A => 'R',
+	R => 'A',
+);
+
+sub undo {
+	my $self = shift;
+	return unless @{$self->{undo}};
+	my $step = pop @{$self->{undo}};
+	push @{$self->{redo}}, [@$step];
+	$$step[0] = $reverse_step{$$step[0]};
+	$self->_do_step(@$step);
+}
+
 sub redo {
+	my $self = shift;
+	return unless @{$self->{redo}};
+	my $step = pop @{$self->{redo}};
+	push @{$self->{undo}}, [@$step];
+	$self->_do_step(@$step);
+}
+
+sub _record_step {
+	my ($self, $step, $begin, $end, @data) = @_;
+	return if $self->{_undo_lock};
+	$self->{_redo} = [];
+	push @{$self->{undo}}, [$step, $begin, $end, @data];
+}
+
+sub _do_step {
+	my ($self, $step, $begin, $end, @data) = @_;
+	$self->{_undo_lock} = 1;
+	$_ = $self->get_iter_at_offset($_) for $begin, $end;
+	if ($step eq 'I') { # insert
+		$self->insert_blocks($begin, @data);
+	}
+	elsif ($step eq 'D') { # delete
+		$self->delete($begin, $end);
+	}
+	elsif ($step eq 'A') { # add tag
+	}
+	elsif ($step eq 'R') { # remove tag
+	}
+	else { die "BUG: unknown step type: $step" }
+	$self->{_undo_lock} = 0;
 }
 
 =item C<copy_clipboard>
@@ -699,6 +791,9 @@ sub _dump_tree { # TODO add "\n" around headings and support lists
 	my $string = '';
 	$string .= ref($_) ? _dump_tree($_) : $_
 		for @$tree[2 .. $#$tree];
+	$string .= "\n" if $$tree[0] =~ /^head/;
+	$string .= "\n"x$$tree[1]{empty_lines}
+		if $$tree[1]{empty_lines};
 	return $string;
 }
 
@@ -733,39 +828,34 @@ sub paste_clipboard {
 		replace     => $replace,
 	};
 	
-	$clipboard->request_contents(
-		Gtk2::Gdk::Atom->intern('ZIM_TREE', 0),
-		\&_clipboard_paste, $paste_data   );
-}
-
-sub _clipboard_paste {
-	# Paste data type negotiating
-	my ($clipboard, $selection, $paste_data) = @_;
-	my $type = $selection->target->name;
-	my $data = $selection->data;
-	#print STDERR '_clipboard_paste type: '.$selection->target->name.", data: $data\n";
-	
-	if ($type eq 'ZIM_TREE' and $_clipboard and $data eq "$_clipboard") {
-		$paste_data->{type} = $type;
-		&_clipboard_paste_data($clipboard, $data, $paste_data);
+	my %targets = map {($_->name => $_)} $clipboard->wait_for_targets();
+	#warn "Available targets: ".join(' ', keys %targets)."\n";
+	my ($atom) = grep defined($_),
+		@targets{qw#ZIM_TREE text/x-zim-page-list text/uri-list#};
+	if ($atom) {
+		$clipboard->request_contents(
+			$atom, \&_clipboard_paste, $paste_data);
 	}
 	else {
-		# Not pasting from a zim widget 
-		# or pasting from a different process
-		$paste_data->{type} = 'STRING';
-		$clipboard->request_text(\&_clipboard_paste_data, $paste_data);
+		$clipboard->request_text(
+			\&_clipboard_paste, $paste_data);
 	}
 }
 
-sub _clipboard_paste_data {
-	# Actual paste
-	my ($clipboard, $string, $paste_data) = @_;
+sub _clipboard_paste { # callback which does the actual paste
+	my ($clipboard, $data, $paste_data) = @_;
 	my $self = $paste_data->{buffer};
-	my $type = $paste_data->{type};
-	#print STDERR "_clipboard_paste_data type: $type, string: $string\n";
+	my $type = ref($data) ? $data->target->name : 'text/plain' ;
+	my $string = ref($data) ? $data->data : $data;
 
 	return unless length $string;
-	return if $type eq 'ZIM_TREE' and "$string" ne "$_clipboard";
+	if ($type eq 'ZIM_TREE' and "$string" ne "$_clipboard") {
+		# not same process - fall back
+		$clipboard->request_text(
+			\&_clipboard_paste, $paste_data);
+		return;
+	}
+	#warn "pasting: $type\n$string\n...\n";
 
 	$self->begin_user_action if $paste_data->{interactive};
 	
@@ -780,23 +870,55 @@ sub _clipboard_paste_data {
 		$iter = $self->get_iter_at_mark( $mark );
 		$self->delete_mark($mark);
 	}
-	else { $iter = $self->get_iter_at_mark( $self->get_insert ) }
 	
 	# actual insert
-	if ($paste_data->{type} eq 'ZIM_TREE') {
-		#use Data::Dumper; print STDERR '_clipboard: ', Dumper $_clipboard;
-		$self->insert_blocks($iter, $_clipboard);
+	my $tree;
+	if ($type eq 'ZIM_TREE') { $tree = $_clipboard }
+	elsif ($type eq 'text/plain') {
+		# FIXME url parsing copy-pasted from Zim/Formats/Wiki.pm
+		my $i = 0;
+		my @tree = ('Page', {},
+			map   { ($i++ % 2) ? _parse_link($_) : $_ }
+			split m#(
+				\b\w[\w\+\-\.]+://  \S*\[\S+\](?:\S+[\w\/])? |
+				\b\w[\w\+\-\.]+://  \S+[\w\/]                |
+				\bmailto:\S+\@      \S*\[\S+\](?:\S+[\w/])?  |
+				\bmailto:\S+\@      \S+[\w/]                 |
+				\b\S+\@\S+\.\w+\b
+			)#x, $string);
+			# The host name in an uri can be "[hex:hex:..]" for ipv6
+			# but we do not want to match "[http://foo.org]"
+			# See rfc3986 for the official -but unpractical- regex
+		$tree = \@tree;
 	}
-	else { # text
-		$paste_data->{interactive}
-			? $self->insert_interactive(
-				$iter, $string, $paste_data->{editable} )
-			: $self->insert($iter, $string) ;
+	else { # text/uri-list or text/x-zim-page-list
+		my @links = Zim::GUI::Component->decode_uri_list($string);
+		if ($type eq 'text/x-zim-page-list') {
+			my $rep = $self->{_zim_name}; # FIXME FIXME FIXME
+			@links = map {s/^\Q$rep\E\?//; $_} @links;
+		}
+		$tree = [ 'Page', {}, 
+			map {(['link', {to => $_}, $_], "\n")} @links ];
 	}
 
+	if ($iter) { $self->insert_blocks($iter, $tree)           }
+	else       { $self->insert_blocks_at_cursor($iter, $tree) }
 	$self->end_user_action if $paste_data->{interactive};
 }
 
+sub _parse_link { # FIXME copy-pasted from Zim/Formats/Wiki.pm
+	my $l = shift;
+	my $text;
+	if ($l =~ /^\[\[([^|]+)\|?(.*)\]\]$/) { # [[link]] or [[link|text]]
+		($l, $text) = ($1, $2);
+	}
+	if ($l =~ /^mailto:|^\S+\@\S+\.\w+$/) {
+		$text = $l unless length $text;
+		$l =~ s/^(mailto:)?/mailto:/;
+	}
+	return ['link', {to => $l}, length($text) ? $text : $l ];
+}
+
 1;
 
 __END__
