@@ -6,10 +6,12 @@ use File::Glob ':glob';
 use File::Spec;
 use File::Copy ();
 use Encode;
+use Zim::Events;
 
 require Cwd;
 
-our $VERSION = '0.16';
+our $VERSION = '0.18';
+our @ISA = qw/Zim::Events/;
 
 use overload
 	'""' => \&path,
@@ -35,6 +37,8 @@ Zim::File - OO wrapper for files
 This package provides an object for handling files.
 It bundles some snippets of code that are needed by several modules.
 
+Inherits from L<Zim::Events>, supports a "write" event.
+
 =head1 OVERLOAD
 
 This class overloads the '""' operator, so the string version of an object
@@ -57,6 +61,22 @@ sub new {
 	bless {path => $class->abs_path($name)}, $class;
 }
 
+=item C<set_check_mtime(BOOLEAN)>
+
+Set whether to check mtime on read / write for this file.
+When turned on an exception is raised when the mtime has
+changed between read and write.
+
+=cut
+
+# feature turned off when undefined
+# when "0" turned on but no mtime set yet
+
+sub set_check_mtime {
+	$_[0]{mtime} = $_[1] ? ($_[0]{mtime} || 0) : undef;
+	$_[0]->signal_connect('write', \&_set_mtime) if $_[1];
+}
+
 =item C<path>
 
 Returns the full path to the file.
@@ -100,13 +120,11 @@ Returns empty when the file does not exist.
 
 sub read {
 	my $self = shift;
-	my $file = $self->{path};
-	return unless -f $file;
-	CORE::open FILE, "<$file" or die "Could not read file: $file\n";
-	binmode FILE, ':utf8' unless $] < 5.008;
-	my @text = wantarray ? (<FILE>) : join('', <FILE>);
-	close FILE;
-	return wantarray ? (@text) : $text[0];
+	return unless $self->exists;
+	my $fh = $self->open('r');
+	my @text = <$fh>;
+	$fh->close;
+	return wantarray ? (@text) : join('', @text);
 }
 
 =item C<read_config()>
@@ -128,14 +146,17 @@ sub read_config {
 	return wantarray ? (\%config, \@order) : \%config;
 }
 
-=item C<read_hash(KEY)>
+=item C<read_hash(KEY, CASE)>
 
 Searches for KEY in a plain text hash.
+The boolean CASE should be set for case sensitive lookup.
+Case insensitive is the default.
 
 =cut
 
 sub read_hash {
-	my ($self, $key) = @_;
+	my ($self, $key, $case) = @_;
+	$key = $case ? quotemeta($key) : qr/(?i:\Q$key\E)/ ;
 	my $value = undef;
 	my $fh = $self->open;
 	while (<$fh>) {
@@ -144,7 +165,7 @@ sub read_hash {
 		$value = $1;
 		last;
 	}
-	close $fh;
+	$fh->close;
 	return $value;
 }
 
@@ -161,24 +182,24 @@ Creates file and parent directories when needed.
 =cut
 
 sub write {
-	my $self = shift;
-	$self->_write('>', $self->{path}, @_);
+	carp "## Zim::File::write: $_[0]";
+	unshift @_, 'w';
+	goto \&_write;
 }
 
 sub append {
-	my $self = shift;
-	$self->_write('>>', $self->{path}, @_);
+	carp "## Zim::File::append: $_[0]";
+	unshift @_, 'a';
+	goto \&_write;
 }
 
 sub _write {
-	my $self = shift;
 	my $mode = shift;
-	my $file = shift;
-	$self->make_dir;
-	CORE::open FILE, $mode.$file or die "Could not write file: $file\n";
-	binmode FILE, ':utf8' unless $] < 5.008;
-	print FILE @_;
-	close FILE or die "Could not write file: $file\n";
+	my $self = shift;
+	$self->make_dir if defined $self->{path};
+	my $fh = $self->open($mode);
+	print $fh @_;
+	$fh->close or die "$!: $self->{path}\n";
 }
 
 =item C<write_config(\%CONFIG, \@ORDER)>
@@ -206,6 +227,7 @@ sub remove {
 	my $self = shift;
 	my $file = $self->{path};
 	unlink $file or die "Could not delete file: $file\n";
+	$self->{mtime} = 0 if defined $self->{mtime};
 	$self->remove_dir;
 }
 
@@ -234,41 +256,92 @@ sub stat {
 
 =item C<open(MODE)>
 
-Returns an L<IO::File> object.
+Returns an C<IO> object that can be used as a file handle.
+
+Make sure to close this handle using OO syntax, using C<< $fh->close >>
+instead of C<close $fh>.
 
 =cut
 
 sub open {
-	my ($self, $mode, $cb) = @_;
+	my ($self, $mode) = @_;
 	$mode ||= 'r';
+	#carp "## Zim::File::open($self, $mode)";
 	my $file = $self->{path};
-	my $handle = IO::File::Zim->new($file, $mode) || die "$file: $!\n";
-	$handle->set_callback($cb) if defined $cb;
-	binmode $handle, ':utf8' if defined $handle and not $] < 5.008;
-		# no utf8 binmode for perl 5.6 and earlier
-		# no OO call because IO::File OO version broken in perl 5.8.7
+	if (defined $self->{mtime}) {
+		my $mtime = (CORE::stat $file)[9];
+		die "File has changed on disk since reading: $file\n"
+			if $mode eq 'w'
+			and $self->{mtime} and $self->{mtime} != $mtime ;
+		$self->{mtime} = $mtime; 
+	}
+	my $handle = IO::File::Zim->new($file, $mode) || die "$!: $file\n";
+	$handle->on_close(sub { $self->signal_emit('write', $file) })
+		if $mode eq 'w';
+	binmode $handle, ':utf8' unless $] < 5.008;
 	return $handle;
 }
 
-=item C<grep(REGEXP)>
+sub _set_mtime { # called by 'write' event
+	$_[0]{mtime} = (CORE::stat $_[0]{path})[9] if defined $_[0]{mtime};
+}
 
-Line based grep function.
+=item C<grep(REGEXP, MODE)>
+
+Line based grep function. MODE can be undefined, 'lines' or 'count';
+In normal mode matches are returned, for 'lines' complete lines that
+match are returns, and for 'count' a number of matches is returned.
 
 =cut
 
 sub grep {
-	my ($self, $regexp) = @_;
+	my ($self, $regexp, $mode) = @_;
 	my $file = $self->{path};
-	return unless -f $file;
-	#warn "Grepping for /$regexp/ in $file\n";
+	$mode ||= 'normal';
+	return $mode eq 'count' ? 0 : () unless -f $file;
+	#print STDERR "Grepping for /$regexp/ in $file (mode: $mode)\n\t";
 	my @match;
-	CORE::open FILE, "<$file" or die "Could not read file: $file\n";
-	while (<FILE>) {
-		push @match, ($_ =~ /($regexp)/);
+	my $fh = $self->open;
+	if ($mode eq 'lines') {
+		while (<$fh>) {
+			push @match, $_ if $_ =~ $regexp;
+			#print STDERR '.';
+		}
 	}
-	close FILE;
-	#warn "\t=> @match\n";
-	return @match;
+	elsif ($mode eq 'count') {
+		$match[0] = 0;
+		while (<$fh>) {
+			$match[0]++ if $_ =~ $regexp;
+			#print STDERR '.';
+		}
+	}
+	else {
+		$regexp  = /($regexp)/;
+		while (<$fh>) {
+			push @match, ($_ =~ $regexp);
+			#print STDERR '.';
+		}
+	}
+	$fh->close;
+	#print STDERR "DONE\n";
+	#print STDERR "\t=> @match\n";
+	return wantarray ? @match : $match[0] ;
+}
+
+=item C<touch()>
+
+Create empty file if file does not exist.
+
+=cut
+
+sub touch {
+	my $file = pop;
+	carp "## Zim::File::touch: $file";
+	return if -e $file;
+	Zim::File->make_dir( Zim::File->dir($file) );
+	CORE::open FH, '>', $file or die "$!: $file";
+	print '';
+	close FH;
 }
 
 =back
@@ -315,6 +388,35 @@ sub abs_path {
 	return $file;
 }
 
+=item C<rel_path(FILE, REFERENCE, UPWARD)>
+
+Returns a path for FILE relative to REFERENCE
+or undef when this is not possible.
+
+When UPWARD is true the returned path can contain
+"../", else it will only go below REFERENCE.
+
+=cut
+
+sub rel_path {
+	my ($class, $file, $base, $upward) = @_;
+	$file = $class->abs_path($file);
+	$base = $class->abs_path($base);
+	my @file = split '/', $file;
+	my @base = split '/', $base;
+	return undef if $base[0] ne $file[0];
+		# might not even be same volume
+	while (@base) {
+		last unless $base[0] eq $file[0];
+		shift @base;
+		shift @file;
+	}
+	return undef if @base and ! $upward;
+	if (@base) { unshift @file, '..' for @base }
+	else       { unshift @file, '.'            }
+	return join '/', @file;
+}
+
 =item C<parse_uri(URI)>
 
 Method for parsing file uris, returns either a normal path or a (new) uri.
@@ -416,10 +518,29 @@ sub list_dir {
 	my @items = grep {! /^\./} readdir DIR;
 	closedir DIR;
 
-	eval {$_ = Encode::decode_utf8($_, 1)} for @items;
+	eval {$_ = Encode::decode('utf8', $_, 1)} for @items;
+	#warn "$dir =>\n", map "\t$_\n", @items;
 	return @items;
 }
 
+=item C<grep_dir(DIR, REGEX)>
+
+Returns all items in DIR that match REGEX.
+
+=cut
+
+sub grep_dir {
+	my ($self, $dir, $regex) = @_;
+	my @matches;
+	opendir DIR, $dir or die "Could not open dir: $dir\n";
+	while (my $item = readdir DIR) {
+		eval { $item = Encode::decode('utf8', $item, 1) };
+		push @matches, $item if $item =~ $regex;
+	}
+	closedir DIR;
+	return @matches;
+}
+
 =item C<make_dir(DIR)>
 
 Create DIR is it doesn't exist already. Subtle differences with L<File::Path>.
@@ -471,6 +592,87 @@ sub remove_dir {
 	}
 }
 
+=item C<resolve_file(\%OPT, $ROOT, @PARTS)>
+
+Look for a file below ROOT that kind of looks like @PARTS.
+Options that can be passed in %OPT include:
+
+=over 4
+
+=item ext
+
+File extension to add to the last part when resolving a file.
+
+If a dir does not exist we try a file with this extension to resolve
+the right case.
+
+=item is_dir
+
+We are looking for a dir.
+
+=back
+
+=cut
+
+sub resolve_file {
+	my ($class, $opt, $path, @parts) = @_;
+	#warn "Resolve file: $path @parts\n";
+
+	while (@parts) {
+		my $part = shift @parts;
+		my $match;
+		if (-d $path) {
+			my $is_file = (! @parts and ! $$opt{is_dir});
+			$match = _grep_match($class, $path, $part, $$opt{ext}, $is_file);
+		}
+		if (defined $match) {
+			#warn "\t$part => $match\n";
+			$path .= '/' . $match;
+		}
+		else {
+			$path = join '/', $path, $part, @parts;
+			$path .= '.' . $$opt{ext} if ! $$opt{is_dir} and length $$opt{ext};
+			last;
+		}
+	}
+
+	#warn "\t=>$path\n";
+	return $path;
+}
+
+sub _grep_match {
+	my ($class, $path, $part, $ext, $is_file) = @_;
+	my $regex = length($ext)
+		? qr/^(?i:\Q$part\E)(?:$|\.$ext$)/
+		: qr/^(?i:\Q$part\E)(?:$|\.)/  ;
+	my @matches = $class->grep_dir($path, $regex);
+	return undef unless @matches;
+
+	if ($is_file and length $ext) { # prefer with extension
+		my $f = $part . '.' . $ext;
+		my @m = grep /\.\Q$ext\E$/, @matches;
+
+		if (@m) {
+			return $f if grep {$_ eq $f} @m;
+			@m = sort @m;
+			return $m[0];
+		}
+		else {
+			return $f if grep {$_ eq $part} @matches;
+			@matches = sort @matches;
+			return $matches[0] . '.' . $ext;
+		}
+	}
+	else {
+		return $part if grep {$_ eq $part} @matches;
+
+		@matches = sort @matches;
+		return substr($matches[0], 0, length $part);
+	}
+
+	return undef;
+}
+
 package IO::File::Zim;
 
 use IO::File;
@@ -482,24 +684,24 @@ our @ISA = qw/IO::File/;
 # stuff like saving the 'mtime' _after_ the file was writen.
 
 # It uses a global hash to save object attributes because
-# IO::File objects are filehandle references, not hashes.
+# IO::Handle objects are filehandle references, not hashes.
 
 # It uses a indirect subroutine definition to circumvent
 # the Pod::Cover test
 
 our %_objects;
 
-*set_callback = sub {
-	my $self = shift;
-	$_objects{"$self"}{callback} = shift;
+*on_close = sub { # used to define a callback
+	my $fh = shift;
+	$_objects{"$fh"}{on_close} = shift;
 };
 
 *close = sub {
-	my $self = shift;
-	my $re = $self->SUPER::close(@_);
-	return $re unless exists $_objects{"$self"};
-	my $obj = delete $_objects{"$self"};
-	$obj->{callback}->() if defined $obj->{callback};
+	my $fh = shift;
+	my $re = $fh->SUPER::close(@_);
+	return $re unless exists $_objects{"$fh"};
+	my $obj = delete $_objects{"$fh"};
+	$obj->{on_close}->() if $obj->{on_close};
 	return $re;
 };
 
