@@ -2,12 +2,12 @@ package Zim::Formats::Wiki;
 
 use strict;
 no warnings;
+use Zim::Formats;
 
-our $VERSION = '0.15';
+our $VERSION = '0.18';
+our @ISA = qw/Zim::Formats/;
 
 # TODO: some tags can be nested: email links for example
-# TODO: header in the middle of a para doesn't get recognized
-
 
 =head1 NAME
 
@@ -20,6 +20,16 @@ It uses a wiki-style syntax to format plain text.
 
 FIXME more verbose description
 
+All format types are signified by double characters,
+this is done to prevent accidental formatting when the same
+characters are used normally in a text. For the same reason
+the strike character is '~' instead of '-' because '--' can
+occur in normal ascii text. In the regexes we try to always
+match the inner pair if more than two of these characters
+are encountered. (Thus if we see C<***bold***> we should match
+C<*(**bold**)*>.)
+
+
 =head1 METHODS
 
 =over 4
@@ -36,6 +46,7 @@ sub load_tree { # TODO whitelines between verbatim blocks should be preserved
 	my @tree;
 	my $para = '';
 	while (<$io>) {
+		s/\r?\n$/\n/; # DOS to Unix conversion
 		if (!/\S/) {
 			push @tree, $class->parse_para($para) if $para =~ /\S/;
 			$tree[-1][1]{empty_lines}++ if scalar @tree;
@@ -46,28 +57,41 @@ sub load_tree { # TODO whitelines between verbatim blocks should be preserved
 			push @tree, $class->parse_head($_);
 			$para = '';
 		}
+		#elsif (/^\$\$+\s+.*?\s+\$\$\s*$/) {
+		#	push @tree, $class->parse_para($para) if $para =~ /\S/;
+		#	push @tree, $class->parse_object($_);
+		#	$para = '';
+		#}
 		else { $para .= $_ }
 	}
 	push @tree, $class->parse_para($para) if $para =~ /\S/;
 	
 	#use Data::Dumper; print STDERR Dumper \@tree;
-	return ['Document', {}, @tree];
+	return ['Page', {}, @tree];
 }
 
 sub parse_head { # parse a header
 	my ($class, $head) = @_;
 	chomp $head;
-	$head =~ s/^(==+)\s+(.*?)(\s+==+|\s*)$/$2/;
+	$head =~ s/^(==+)\s+(.*?)(\s+==+\s*|\s*)$/$2/;
 	my $level = 7 - length($1); # =X6 => head1, =X5 => head2 etc.
 	$level = 1 if $level < 1;
 	return ['head'.$level, {}, $head];
 }
 
-sub parse_para {
+sub parse_object { # FIXME use image syntax for all objects
+	my ($class, $obj) = @_;
+	chomp $obj;
+	$obj =~ s/^\$\$\s+(.*?)\s+\$\$\s*$/$1/;
+	return ['object', {type => 'text'}, $obj];
+}
+
+sub parse_para { # distinguise normal paras from verbatim paras
 	my ($class, $text) = @_;
-	return ( ($text =~ /^(?!\t|\s\s+)/s) # one tab or two spaces
-		? ['Para', {}, $class->parse_block($text)]
-		: $class->parse_Verbatim($text)            );
+	return  $class->parse_Verbatim($text) unless $text =~ /^(?!\t|\s\s+)/m;
+	#$text =~ s/^\.//mg unless $text =~ /^(?!\.?\t|\.?\s\s+)/s;
+	$text =~ s/^(\s*)\*(\s+)/$1\x{2022}$2/mg;
+	return ['Para', {}, $class->parse_block($text)];
 }
 
 sub parse_Verbatim { # parse pre formated paragraphs
@@ -77,7 +101,7 @@ sub parse_Verbatim { # parse pre formated paragraphs
 	return ['Verbatim', {}, $pre];
 }
 
-our @parser_subs = qw/parse_verbatim parse_link parse_styles parse_image/;
+our @parser_subs = qw/parse_verbatim parse_links parse_images parse_styles parse_urls/;
 
 sub parse_block { # parse a block of text
 	my ($class, @text) = @_;
@@ -93,33 +117,42 @@ sub parse_verbatim { # like parse_style() but needs to be done earlier
 	return	map {
 		unless ($i++ % 2) { $_ }
 		elsif (/^\'\'(.+)\'\'$/) { ['verbatim',  {}, $1] }
-	} split /(\'\'.+?\'\')/, $text;
+	} split /(\'\'(?!\').+?\'\')/, $text;
 }
 
-sub parse_link { # parse links and urls
+sub parse_links {
 	my ($class, $text) = @_;
 	my $i = 0;
-	return map {
-		if ($i++ % 2) {
-			my $text;
-			if (s/^\[\[(.+)\|(.+)\]\]$/$1/) { $text = $2 }
-			else { s/^\[\[(.+)\]\]$/$1/ }
-			
-			if (/^mailto:|^\S+\@\S+\.\w+$/) {
-				$text = $_ unless length $text;
-				$_ =~ s/^(mailto:)?/mailto:/;
-			}
-			
-			['link', {to => $_}, length($text) ? $text : $_ ]
-		} else { $_ }
-	} split m#(
-		\[\[.+?\]\]                                  |
-		\b\w[\w\+\-\.]+://  \S*\[\S+\](?:\S+[\w\/])? |
-		\b\w[\w\+\-\.]+://  \S+[\w\/]                |
-		\bmailto:\S+\@      \S*\[\S+\](?:\S+[\w/])?  |
-		\bmailto:\S+\@      \S+[\w/]                 |
-		\b\S+\@\S+\.\w+\b
-	)#x, $text;
+	return
+		map   { ($i++ % 2) ? _parse_link($_) : $_ }
+		split m#(\[\[(?!\[).+?\]\])#, $text;
+}
+
+sub _parse_link {
+	my $l = shift;
+	my $text;
+	if ($l =~ /^\[\[([^|]+)\|?(.*)\]\]$/) { # [[link]] or [[link|text]]
+		($l, $text) = ($1, $2);
+	}
+	if ($l =~ /^mailto:|^\S+\@\S+\.\w+$/) {
+		$text = $l unless length $text;
+		$l =~ s/^(mailto:)?/mailto:/;
+	}
+	return ['link', {to => $l}, length($text) ? $text : $l ];
+}
+
+sub parse_urls {
+	my ($class, $text) = @_;
+	my $i = 0;
+	return
+		map   { ($i++ % 2) ? _parse_link($_) : $_ }
+		split m#(
+			\b\w[\w\+\-\.]+://  \S*\[\S+\](?:\S+[\w\/])? |
+			\b\w[\w\+\-\.]+://  \S+[\w\/]                |
+			\bmailto:\S+\@      \S*\[\S+\](?:\S+[\w/])?  |
+			\bmailto:\S+\@      \S+[\w/]                 |
+			\b\S+\@\S+\.\w+\b
+		)#x, $text;
 		# The host name in an uri can be "[hex:hex:..]" for ipv6
 		# but we do not want to match "[http://foo.org]"
 		# See rfc3986 for the official -but unpractical- regex
@@ -135,19 +168,27 @@ sub parse_styles { # parse blocks of bold, italic and underline
 		elsif (/^\~\~(.+)\~\~$/) { ['strike',    {}, $1] }
 		elsif (/^\_\_(.+)\_\_$/) { ['underline', {}, $1] }
 	} split /(
-		\*\*.+?\*\* |
-		(?<!\:)\/\/.+?\/\/ |
-		\~\~.+?\~\~ |
-		__.+?__
+		\*\*(?!\*).+?\*\* |
+		(?<!\:)\/\/(?!\/).+?\/\/ |
+		\~\~(?!\~).+?\~\~ |
+		__(?!_).+?__
 	)/x, $text;
 }
 
-sub parse_image {
+sub parse_images {
 	my ($class, $text) = @_;
 	my $i = 0;
-	return map {
-		($i++ % 2) ? ['image', {src => $_}] : $_
-	} split /\{\{(.+?)\}\}/, $text;
+	my @parts =
+		map   { ($i++ % 2) ? ['image', {src => $_}] : $_ }
+		split /\{\{(?!\{)(.+?)\}\}/, $text;
+	for my $p (@parts) {
+		next unless ref $p and $$p[0] eq 'image';
+		if ($$p[1]{src} =~ s/\?(\w+=\w+(?:&\w+=\w+)*)$//) {
+			my %arg = split /[=&]/, $1;
+			$$p[1]{$_} = $arg{$_} for keys %arg;
+		}
+	}
+	return @parts;
 }
 
 =item C<save_tree(IO, TREE, PAGE)>
@@ -177,6 +218,7 @@ sub _save_tree {
 	while (@$tree) {
 		my $node = shift @$tree;
 		unless (ref $node) {
+			$node =~ s/^(\s*)\x{2022}(\s+)/$1*$2/mg;
 			print $node;
 			next;
 		}
@@ -198,8 +240,14 @@ sub _save_tree {
 			print ( ('='x$n)." $text ".('='x$n)."\n" );
 		}
 		elsif ($tag eq 'image') {
-			print '{{'.$$meta{src}.'}}';
+			my $file = $$meta{src};
+			my @k = grep $$meta{$_}, qw/height width/;
+			$file .= '?' . join '&', map "$_=$$meta{$_}", @k if @k;
+			print '{{'.$file.'}}';
 		}
+#		elsif ($tag eq 'object') { # FIXME use image syntax for all objects
+#			print "\$\$ $text \$\$\n";
+#		}
 		elsif ($tag eq 'link') {
 			my $to = $$meta{to};
 			$to =~ s/^mailto:// unless $text =~ /^mailto:/;
@@ -254,7 +302,6 @@ modify it under the same terms as Perl itself.
 =head1 SEE ALSO
 
 L<Zim>,
-L<Zim::Repository>,
 L<Zim::Page>
 
 =cut
